import logging
from typing import ClassVar

__version__ = "1.0.0"
__repo_url__ = "https://github.com/sakurazaki/discord-continued"


class Data:
    """An object representing constants for the library."""

    LOGGER: ClassVar[int] = logging.DEBUG
